#include <stdio.h>
#include <float.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#define BUFFER_SIZE  200
#define NUM_CELL_TYPES  8
#define MAGIC 0x7
#define MAGIC_V3 0x5

#include "ucd.h"
#include "l3volum.h"

#ifndef LINT
    static const char sccsid[] = "%W% %G% Paul Letexier";
#endif /* !LINT */

#define MALLOC(type,size) (type*) malloc(((size) == 0 ? 1 : (size))* \
				sizeof(type))
#define FREE(ptr) if (ptr) free(ptr);ptr = NULL

#if defined(__STDC__)
#  define VOID void*
# else
#  define VOID char*
#endif

static char* cell_list[NUM_CELL_TYPES] = 
            {"pt", "line", "tri", "quad", "tet", "pyr", "prism", "hex"};

static int  num_node[NUM_CELL_TYPES] = { 1, 2, 3, 4, 4, 5, 6, 8 };

/*
 *------------------------------------------------------------------------*
 *                                                                        *
 *     ****  get_token : returns the next value in tline[]      ****      *
 *                        (separated by a blank)                          *
 *                                                                        *
 *------------------------------------------------------------------------*/

static int
get_token(char** tline,VOID data,char* control)
{
	char* line;
	char token[80];
	char c;

	int ntc;

	/**************
	 ***  body  ***
	 **************/

	c = 0;
	line = *tline;

	for (ntc = 0; c != '\n'; line++)
	{
	    c  = *line;

	    assert( ntc <= (int) strlen(*tline) );

	    if ((c != ' ') && (c != '\n'))
	      /*    c est un bon caractere */
	        token[ntc++] = c;
	    else if (ntc != 0)
	    {
		/*    c est un delimiteur terminal de token, on a termine */
		token[ntc] = '\0';
		sscanf (token, control, data);
		*tline = line;
		break;
	    }
#if 0
            else
  /*    c est un delimiteur initial de token, on cherche un bon caractere */
            ;
#endif
	}

	return (ntc);
}

/*
 *------------------------------------------------------------------------*
 *                                                                        *
 *   ****  check_file : check to see if the file to read is binary ****   *
 *                                                                        *
 *------------------------------------------------------------------------*/

static int check_file (FILE* file_num)
{ 
	int ret = 0;

	/**************
	 ***  body  ***
	 **************/

	if (!file_num)
	{
	    fprintf(stderr," Error in read_ucd: can't read file %p\n",file_num);
	    exit(1);
	}
	else
	{
	    char c = fgetc(file_num); 

	    if (c != MAGIC)
	    {
		ret = 1;
		if (c == MAGIC_V3) fprintf(stderr,
			 " Sorry it is an AVS3 binary file !\n");
	    }

	    rewind(file_num);
	}

	return(ret);
}

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *      ****  read_data : read a block of data from an ascii file  ****     *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static void read_data (FILE* fp,int num_entity,float* data,
                       float* min_data,float* max_data,char* labels,
                       char* units,int* num_comp,int** pcomp_list)
{

  char line[BUFFER_SIZE];

  int entity, comp, n, m;
  int *comp_list ;

 /**************
  ***  body  ***
  **************/

  /*  read the number of components */

  fscanf (fp, "%d", num_comp);

  comp_list = MALLOC( int,*num_comp);
  *pcomp_list= comp_list;

  /*  read dimension of each componant */

  for (comp = 0; comp < *num_comp; comp++) fscanf (fp, "%d", &comp_list[comp]);

  /*  read the component names and units.  */

  fgets (line, BUFFER_SIZE, fp);

  for (comp = 0, n = 0, m = 0; comp < *num_comp; comp++)
  {
    int j;

    fgets (line, BUFFER_SIZE, fp);

    /* scan label of componant "i" */

    for (j = 0; line[j] == ' '; j++);

    for (; line[j] != ','; j++)
    {
      assert( n <= (int) strlen(labels) );

      labels[n++] = line[j];
    }

    /* scan unit of componant "i" */

    for (++j; line[j] == ' '; j++);

    for (; line[j] != '\n'; j++)
    {
      assert( m <= (int) strlen(units) );

      units[m++] = line[j];
    }

    /* is there an other componant ? */

    if (comp != *num_comp - 1) 
    {
      assert( n <= (int) strlen(labels) );
      assert( m <= (int) strlen(units) );

      labels[n++] = '.';
      units[m++] = '.';
    }
  }

  assert( n <= (int) strlen(labels) );
  assert( m <= (int) strlen(units) );

  labels[n] = '\0';
  units[m] = '\0';

  for (comp = 0; comp < *num_comp; comp++) min_data[comp] = max_data[comp] = 0.0;

  /* read data and compute min/max */

  for (entity = 0; entity < num_entity; entity++)
  {
    int i;
    int offset = 0;
    int id;
    
    fscanf (fp, "%d", &id);
    assert( (1 <= id)&&(id <= num_entity) );

    for (i = 0; i < *num_comp; i++)
    {
      float value = 0.;
      int j;
      int num_vals = comp_list[i];

      for (j = 0; j < num_vals; j++)
      {
        fscanf (fp, "%f",  &value);
        data[(id-1)*num_vals + offset + j] = value;
      }

      /*  only set max/min for scalar data.  */

      if (num_vals == 1)
      {

        if (entity == 0) min_data[i] = max_data[i] = value;
        else
	{

	  if (min_data[i] > value) min_data[i] = value;

	  if (max_data[i] < value) max_data[i] = value;

	}

      }

      offset += num_entity * num_vals;
    }

  }

}

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *           ****  read_ascii :  read a ucd ascii file. ****                *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static int read_ascii (FILE* file_num,Ctype** pcells,Ntype** pcell_nlists,
                       Mtype* model,float** pxc,float** pyc,float** pzc,
                       float** pnode_data,float**  pcell_data,
                       float**  pmodel_data,
                       float** pmin_node_data,float**  pmax_node_data, 
	               float** pmin_cell_data,float**  pmax_cell_data,
                       float** pmin_model_data,float**  pmax_model_data)
{
  char line[BUFFER_SIZE];

  FILE* fp = file_num;

  int num_nodes, num_cells;
  int num_node_data, num_cell_data, num_model_data;
  int num_nlist_nodes;

  Ctype *cells;
  Ntype *cell_nlists;

  float *xc, *yc, *zc;

  int node, cell, i; 

 /**************/
 /***  body  ***/
 /**************/

  printf("UCD : read ascii\n");

  /*  skip comments.  */

  fgets (line, BUFFER_SIZE, fp);

  if(feof(fp)) return 0;

  while (line[0] == '#') fgets (line, BUFFER_SIZE, fp);

  /*  read model stats.  */ 

  sscanf (line, "%d %d %d %d %d", &num_nodes, &num_cells, &num_node_data, 
          &num_cell_data, &num_model_data);

  xc = MALLOC(float,num_nodes);
  yc = MALLOC(float,num_nodes);
  zc = MALLOC(float,num_nodes);

  /*  read nodal coordinates.  */

  for (node = 0; node < num_nodes; node++)
  {
    int   node_id;
    float coord[3];

    fscanf (fp, "%d %f %f %f", &node_id, &coord[0], &coord[1], &coord[2]);

    node_id = node_id - 1;

    assert( (0 <= node_id)&&(node_id < num_nodes) );
    xc[node_id] = coord[0];
    yc[node_id] = coord[1];
    zc[node_id] = coord[2];
  }
  
  num_nlist_nodes = 0;
  cells           = MALLOC(Ctype,num_cells);
  cell_nlists     = MALLOC(Ntype,num_cells);

  /*  read cell type, topology.  */
  
  fgets (line, BUFFER_SIZE, fp);
  
  for (cell = 0; cell < num_cells; cell++)
  {
    int n = 0;
    char *tl = line;
    int ucd_cell_type;
    char cell_type[10];
    int nc = 0;

    fgets (line, BUFFER_SIZE, fp);
  
    get_token (&tl,(VOID) &cells[cell].id, "%d");
    get_token (&tl,(VOID) &cells[cell].mat, "%d");
  
    nc = get_token (&tl, cell_type, "%s");
    assert( nc < 10 );                       /* nc = MIN( 10, nc ); */
    cell_type[nc] = '\0';

    for (ucd_cell_type = 0; ucd_cell_type < NUM_CELL_TYPES; ucd_cell_type++)
      if (!strcmp(cell_list[ucd_cell_type], cell_type))
        break;

    if (ucd_cell_type == NUM_CELL_TYPES) 
    {
      fprintf(stderr," Error in read_ucd: unknown cell type : %s\n", cell_type);
      cells[cell].cell_type = 0;
    }
    else
    {
      cells[cell].cell_type = ucd_cell_type;
    }

/* n = cell's number of nodes. */

    cell_nlists[cell].node_list = MALLOC( int,num_node[cells[cell].cell_type] );

/* bug corrige le 30/9/93 pour mailles triangulaires il pouvait
   y avoir un 4eme point avec la valeur zero qui etait lu
   et ajoute dans une taille allouee de trois !!             */

    while((n < num_node[cells[cell].cell_type]) &&
	get_token(&tl,(VOID) &cell_nlists[cell].node_list[n++], "%d"));

    cell_nlists[cell].n = cells[cell].n = n;
    num_nlist_nodes += n;
  }

  /*  read nodal data.  */
  
  if (num_node_data)
  {
    float *node_data     = MALLOC(float, num_nodes * num_node_data);
    float *min_node_data = MALLOC(float, num_node_data);
    float *max_node_data = MALLOC(float, num_node_data);

    read_data (fp, num_nodes, node_data, min_node_data, max_node_data, 
	       model->node_data_labels, model->node_data_units,
	       &model->num_node_comp, &model->node_comp_list);

  /* ??? It's a bit misterious ... ??? */

    model->node_active_list = MALLOC( int,model->num_node_comp );
    model->node_active_list[0] = 1;

    for (i = 1; i < model->num_node_comp; i++)
      model->node_active_list[i] = 0;
  
    *pnode_data     = node_data;
    *pmax_node_data = max_node_data;
    *pmin_node_data = min_node_data;
  }
  
  /*  read cell data.  */
  
  if (num_cell_data)
  {
    float *cell_data     = MALLOC(float, num_cells * num_cell_data);
    float *min_cell_data = MALLOC(float, num_cell_data);
    float *max_cell_data = MALLOC(float, num_cell_data);
  
    read_data (fp, num_cells, cell_data, min_cell_data, max_cell_data, 
	       model->cell_data_labels, model->cell_data_units,
	       &model->num_cell_comp, &model->cell_comp_list);

  /* ??? It's a bit misterious ... ??? */

    model->cell_active_list = MALLOC( int,model->num_cell_comp );
    model->cell_active_list[0] = 1;

    for (i = 1; i < model->num_cell_comp; i++)
      model->cell_active_list[i] = 0;
  
    *pcell_data     = cell_data;
    *pmin_cell_data = min_cell_data;
    *pmax_cell_data = max_cell_data;
  }
  
  /*  read and store model data.  */
  
  if (num_model_data)
  {
    float *model_data     = MALLOC(float, num_model_data);
    float *min_model_data = MALLOC(float, num_model_data);
    float *max_model_data = MALLOC(float, num_model_data);
  
    read_data (fp, num_model_data, model_data, min_model_data, max_model_data, 
	       model->model_data_labels, model->model_data_units,
	       &model->num_model_comp, &model->model_comp_list);

  /* ??? It's a bit misterious ... ??? */

    model->model_active_list = MALLOC( int,model->num_model_comp );
    model->model_active_list[0] = 1;

    for (i = 1; i < model->num_model_comp; i++)
      model->model_active_list[i] = 0;

    *pmodel_data     = model_data;
    *pmax_model_data = max_model_data;
    *pmin_model_data = min_model_data;
  }

  *pxc = xc, *pyc = yc, *pzc = zc;
  *pcells = cells;
  *pcell_nlists = cell_nlists;

  model->num_nodes       = num_nodes;
  model->num_cells       = num_cells;
  model->num_nlist_nodes = num_nlist_nodes;
  model->num_node_data   = num_node_data;
  model->num_cell_data   = num_cell_data;
  model->num_model_data  = num_model_data;

  return(1);
}

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *          ****  read_binary :  read a ucd binary file. ****               *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static int read_binary (FILE* file_num,Ctype** pcells,int** pcell_nlists,
                        Mtype* model,float** pxc,float**  pyc,float**  pzc,
                        float** pnode_data,float**  pcell_data,
                        float** pmodel_data,
                        float** pmin_node_data,float** pmax_node_data,
                        float** pmin_cell_data,float** pmax_cell_data,
                        float** pmin_model_data,float** pmax_model_data)
{
  char magic;
  FILE* fp = file_num;

  int num_nodes, num_cells;
  int num_node_data, num_cell_data, num_model_data;
  int num_nlist_nodes;

  Ctype *cells;
  int   *cell_nlists;

  float *xc, *yc, *zc;

 /**************
  ***  body  ***
  **************/

  printf("UCD : read binary\n");

  fread ((char *)&magic, sizeof(char), 1, fp);
  assert( magic == MAGIC );

  {
    int num[6];

    fread ((char *)num, sizeof(int), 6, fp);

    num_nodes       = num[0];
    num_cells       = num[1];
    num_node_data   = num[2];
    num_cell_data   = num[3];
    num_model_data  = num[4];
    num_nlist_nodes = num[5];
  }

  model->num_nodes       = num_nodes;
  model->num_cells       = num_cells;
  model->num_node_data   = num_node_data;
  model->num_cell_data   = num_cell_data;
  model->num_model_data  = num_model_data;
  model->num_nlist_nodes = num_nlist_nodes;

  cells = MALLOC( Ctype,num_cells );
  cell_nlists = MALLOC( int,num_nlist_nodes );

  fread  ((char *)cells, sizeof(Ctype), num_cells, fp);
  fread  ((char *)cell_nlists, sizeof(int), num_nlist_nodes, fp);

  xc = MALLOC( float,num_nodes );
  yc = MALLOC( float,num_nodes );
  zc = MALLOC( float,num_nodes );

  fread ((char *)xc, sizeof(float), num_nodes, fp);
  fread ((char *)yc, sizeof(float), num_nodes, fp);
  fread ((char *)zc, sizeof(float), num_nodes, fp);

  if (num_node_data)
  {
    char  *node_data_labels = model->node_data_labels;
    char  *node_data_units = model->node_data_units;
    int   num_node_comp;
    int   *node_comp_list   ;
    float *min_node_data    = MALLOC(float, num_node_data);
    float *max_node_data    = MALLOC(float, num_node_data);
    float *node_data        = MALLOC(float, num_nodes * num_node_data);
    int   *node_active_list ;

    fread (node_data_labels, LLAB, 1, fp);
    fread (node_data_units, LLAB, 1, fp);
    fread ((char *)&num_node_comp, sizeof(int), 1, fp);
    node_comp_list   = MALLOC( int,num_node_comp );
    node_active_list = MALLOC( int,num_node_comp );
    fread ((char *)node_comp_list, sizeof(int), num_node_comp, fp);
    fread ((char *)min_node_data, sizeof(float), num_node_data, fp);
    fread ((char *)max_node_data, sizeof(float), num_node_data, fp);
    fread ((char *)node_data, sizeof(float), num_node_data * num_nodes, fp);
    fread ((char *)node_active_list, sizeof(int), num_node_comp, fp);

    model->num_node_comp    = num_node_comp;
    model->node_comp_list   = node_comp_list;
    *pmin_node_data         = min_node_data;
    *pmax_node_data         = max_node_data;
    *pnode_data             = node_data;
    model->node_active_list = node_active_list;
  }

  if (num_cell_data)
  {
    char  *cell_data_labels = model->cell_data_labels;
    char  *cell_data_units = model->cell_data_units;
    int   num_cell_comp;
    int   *cell_comp_list   ;
    float *min_cell_data    = MALLOC(float, num_cell_data);
    float *max_cell_data    = MALLOC(float, num_cell_data);
    float *cell_data        = MALLOC(float, num_cells * num_cell_data);
    int   *cell_active_list ;

    fread (cell_data_labels, LLAB, 1, fp);
    fread (cell_data_units, LLAB, 1, fp);
    fread ((char *)&num_cell_comp, sizeof(int), 1, fp);
    cell_comp_list   = MALLOC( int,num_cell_comp );
    cell_active_list = MALLOC( int,num_cell_comp );
    fread ((char *)cell_comp_list, sizeof(int), num_cell_comp, fp);
    fread ((char *)min_cell_data, sizeof(float), num_cell_data, fp);
    fread ((char *)max_cell_data, sizeof(float), num_cell_data, fp);
    fread ((char *)cell_data, sizeof(float), num_cell_data * num_cells, fp);
    fread ((char *)cell_active_list, sizeof(int), num_cell_comp, fp);

    model->num_cell_comp    = num_cell_comp;
    model->cell_comp_list   = cell_comp_list;
    *pmin_cell_data         = min_cell_data;
    *pmax_cell_data         = max_cell_data;
    *pcell_data             = cell_data;
    model->cell_active_list = cell_active_list;
  }

  if (num_model_data)
  {
    char  *model_data_labels = model->model_data_labels;
    char  *model_data_units  = model->model_data_units;
    int   num_model_comp;
    int   *model_comp_list   ;
    float *min_model_data    = MALLOC(float, num_model_data);
    float *max_model_data    = MALLOC(float, num_model_data);
    float *model_data        = MALLOC(float, num_model_data);
    int   *model_active_list ;

    fread (model_data_labels, LLAB, 1, fp);
    fread (model_data_units, LLAB, 1, fp);
    fread ((char *)&num_model_comp, sizeof(int), 1, fp);
    model_comp_list   = MALLOC( int,num_model_comp );
    model_active_list = MALLOC( int,num_model_comp );
    fread ((char *)model_comp_list, sizeof(int), num_model_comp, fp);
    fread ((char *)min_model_data, sizeof(float), num_model_data, fp);
    fread ((char *)max_model_data, sizeof(float), num_model_data, fp);
    fread ((char *)model_data, sizeof(float), num_model_data, fp);
    fread ((char *)model_active_list, sizeof(int), num_model_comp, fp);

    model->num_model_comp    = num_model_comp;
    model->model_comp_list   = model_comp_list;
    *pmin_model_data         = min_model_data;
    *pmax_model_data         = max_model_data;
    *pmodel_data             = model_data;
    model->model_active_list = model_active_list;
  }

  *pxc = xc, *pyc = yc, *pzc = zc;
  *pcells = cells;
  *pcell_nlists = cell_nlists; 

  return(1);
}

/*
 *------------------------------------------------------------------------*
 *                                                                        *
 *   ****  write_data : print a block of data from an ascii file  ****    *
 *                                                                        *
 *------------------------------------------------------------------------*/

static int write_data(FILE* fp,int num_data,float data[],float min_data[],
                  float max_data[],char data_labels[],char units[],
                  int num_comp,int comp_list[])
{
  int i;
  int node;
  char **Label,**Unit;

  fprintf(fp,"%d",num_comp);

  for (i=0; i<num_comp; i++) fprintf(fp," %d",comp_list[i]);

  fprintf(fp,"\n");

/* Le caractere '.' est le delimiteur des num_comp textes */
  Label=MALLOC(char *,num_comp);
  Label[0]=strtok(data_labels,".");

  for (i=1; i<num_comp; i++) Label[i]=strtok(NULL,".");

  Unit=MALLOC(char *,num_comp);
  Unit[0]=strtok(units,".");

  for (i=1; i<num_comp; i++) Unit[i]=strtok(NULL,".");

  for (i=0; i<num_comp; i++)
  {
  /* On peut avoir des textes vides, on a alors des delimiteurs consecutifs
     strtok rend donc un pointeur NULL; on imprime un blanc. */
    fprintf(fp,"%s, %s\n", (Label[i]?Label[i]:" "), (Unit[i]?Unit[i]:" ") );
  }

  FREE(Label);
  FREE(Unit);

  for (node=0; node<num_data; node++)
  {
    int offset=0;

    fprintf(fp,"%d",node+1);

    for (i=0; i<num_comp; i++)
    {
      int j;

      int num_vals = comp_list[i];

      for (j=0; j<num_vals; j++)
		fprintf(fp," %e",data[node*num_vals + offset + j]);

      offset += num_data * num_vals;
    }

    fprintf(fp,"\n"); /* je n'en suis pas sur */
  }

  return 0;
}

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *           ****  write_ascii :  print a ucd ascii file. ****              *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static int
write_ascii (FILE* file_num,Ctype* cells,Ntype* cell_nlists,
                    Mtype model_stats,float xc[],float  yc[],float  zc[],
                    float node_data[],float  cell_data[],float  model_data[],
                    float min_node_data[],float  max_node_data[],
                    float min_cell_data[],float  max_cell_data[],
                    float min_model_data[],float  max_model_data[])
{
  int node,cell;
  FILE* fp = file_num;

  fprintf(fp,"%d %d %d %d %d\n", model_stats.num_nodes, model_stats.num_cells,
    model_stats.num_node_data, model_stats.num_cell_data,
    model_stats.num_model_data);

  for (node=0; node<model_stats.num_nodes; node++ )
  {
    fprintf(fp,"%d %e %e %e\n", node+1, xc[node] ,yc[node] ,zc[node]);
    }

  for (cell=0; cell<model_stats.num_cells; cell++ )
  {
    int i;

    fprintf(fp,"%d %d %s", cells[cell].id, cells[cell].mat,
                       cell_list[ cells[cell].cell_type ]);

    for (i=0; i<cells[cell].n; i++)
	fprintf(fp," %d",cell_nlists[cell].node_list[i]);

    fprintf(fp,"\n");
  }

  if (model_stats.num_node_data) write_data(fp,
    model_stats.num_nodes, node_data, min_node_data, max_node_data,
    model_stats.node_data_labels,  model_stats.node_data_units,
    model_stats.num_node_comp, model_stats.node_comp_list);

  if (model_stats.num_cell_data) write_data(fp,
    model_stats.num_cells, cell_data, min_cell_data, max_cell_data,
    model_stats.cell_data_labels,  model_stats.cell_data_units,
    model_stats.num_cell_comp, model_stats.cell_comp_list);

  if (model_stats.num_model_data) write_data(fp,
    model_stats.num_model_data, model_data, min_model_data, max_model_data,
    model_stats.model_data_labels,  model_stats.model_data_units,
    model_stats.num_model_comp, model_stats.model_comp_list);

  return 0;
}

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *         ****  write_binary :  write a ucd binary file. ****              *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static int
write_binary (FILE* file_num,Ctype* cells,int* cell_nlists,
                     Mtype model,float xc[],float  yc[],float  zc[],
                     float node_data[],float cell_data[],float model_data[],
                     float min_node_data[],float max_node_data[],
                     float min_cell_data[],float max_cell_data[],
                     float min_model_data[],float max_model_data[])
{
  char magic = MAGIC;
  FILE* fp = file_num;

  int num_nodes, num_cells;
  int num_node_data, num_cell_data, num_model_data;
  int num_nlist_nodes;

 /**************/
 /***  body  ***/
 /**************/

  fwrite (&magic, sizeof(char), 1, fp);

  num_nodes       = model.num_nodes       ;
  num_cells       = model.num_cells       ;
  num_node_data   = model.num_node_data   ;
  num_cell_data   = model.num_cell_data   ;
  num_model_data  = model.num_model_data  ;
  num_nlist_nodes = model.num_nlist_nodes ;

  {
    int num[6];

    num[0] = num_nodes       ;
    num[1] = num_cells       ;
    num[2] = num_node_data   ;
    num[3] = num_cell_data   ;
    num[4] = num_model_data  ;
    num[5] = num_nlist_nodes ;

    fwrite ((char *)num, sizeof(int), 6, fp);
  }

  fwrite  ((char *)cells, sizeof(Ctype), num_cells, fp);
  fwrite  ((char *)cell_nlists, sizeof(int), num_nlist_nodes, fp);

  fwrite ((char *)xc, sizeof(float), num_nodes, fp);
  fwrite ((char *)yc, sizeof(float), num_nodes, fp);
  fwrite ((char *)zc, sizeof(float), num_nodes, fp);

  if (num_node_data)
  {
    char  *node_data_labels = model.node_data_labels;
    char  *node_data_units  = model.node_data_units;
    int   num_node_comp     = model.num_node_comp;
    int   *node_comp_list   = model.node_comp_list;
    int   *node_active_list = model.node_active_list;

    fwrite (node_data_labels, LLAB, 1, fp);
    fwrite (node_data_units, LLAB, 1, fp);
    fwrite ((char *)&num_node_comp, sizeof(int), 1, fp);
    fwrite ((char *)node_comp_list, sizeof(int), num_node_comp, fp);
    fwrite ((char *)min_node_data, sizeof(float), num_node_data, fp);
    fwrite ((char *)max_node_data, sizeof(float), num_node_data, fp);
    fwrite ((char *)node_data, sizeof(float), num_node_data * num_nodes, fp);
    fwrite ((char *)node_active_list, sizeof(int), num_node_comp, fp);
  }

  if (num_cell_data)
  {
    char  *cell_data_labels = model.cell_data_labels;
    char  *cell_data_units  = model.cell_data_units;
    int   num_cell_comp     = model.num_cell_comp;
    int   *cell_comp_list   = model.cell_comp_list;
    int   *cell_active_list = model.cell_active_list;

    fwrite (cell_data_labels, LLAB, 1, fp);
    fwrite (cell_data_units, LLAB, 1, fp);
    fwrite ((char *)&num_cell_comp, sizeof(int), 1, fp);
    fwrite ((char *)cell_comp_list, sizeof(int), num_cell_comp, fp);
    fwrite ((char *)min_cell_data, sizeof(float), num_cell_data, fp);
    fwrite ((char *)max_cell_data, sizeof(float), num_cell_data, fp);
    fwrite ((char *)cell_data, sizeof(float), num_cell_data * num_cells, fp);
    fwrite ((char *)cell_active_list, sizeof(int), num_cell_comp, fp);
  }

  if (num_model_data)
  {
    char  *model_data_labels = model.model_data_labels;
    char  *model_data_units  = model.model_data_units;
    int   num_model_comp     = model.num_model_comp;
    int   *model_comp_list   = model.model_comp_list;
    int   *model_active_list = model.model_active_list;

    fwrite (model_data_labels, LLAB, 1, fp);
    fwrite (model_data_units, LLAB, 1, fp);
    fwrite ((char *)&num_model_comp, sizeof(int), 1, fp);
    fwrite ((char *)model_comp_list, sizeof(int), num_model_comp, fp);
    fwrite ((char *)min_model_data, sizeof(float), num_model_data, fp);
    fwrite ((char *)max_model_data, sizeof(float), num_model_data, fp);
    fwrite ((char *)model_data, sizeof(float), num_model_data, fp);
    fwrite ((char *)model_active_list, sizeof(int), num_model_comp, fp);
  }

  return(1);
}

/*  debut de la declaration des informations lues sur le fichier UCD   */

static  Ctype* cells = NULL;

static  Ntype* cell_nlists = NULL;

static  Mtype model_stats;

static  float *xc = NULL, *yc = NULL, *zc = NULL;
static  float *node_data = NULL,*cell_data = NULL,*model_data = NULL;
static  float *min_node_data = NULL, *max_node_data = NULL;
static  float *min_cell_data = NULL, *max_cell_data = NULL;
static  float *min_model_data = NULL, *max_model_data = NULL;

void
purgeucd(void)
{
	cells = NULL;

	cell_nlists = NULL;

	model_stats.node_comp_list = NULL;
	model_stats.node_active_list = NULL;
	model_stats.cell_comp_list = NULL;
	model_stats.cell_active_list = NULL;
	model_stats.model_comp_list = NULL;
	model_stats.model_active_list = NULL;

	model_stats.num_nlist_nodes = 0;

	xc = NULL, yc = NULL, zc = NULL;
	node_data = NULL,cell_data = NULL,model_data = NULL;
	min_node_data = NULL, max_node_data = NULL;
	min_cell_data = NULL, max_cell_data = NULL;
	min_model_data = NULL, max_model_data = NULL;
}

int
freeucd(void)
{
  int i = 0;

  FREE(cells);

  for(i=0;i<model_stats.num_cells;i++) {FREE(cell_nlists[i].node_list);};

  FREE(cell_nlists);

  FREE(xc);
  FREE(yc);
  FREE(zc);

  FREE(min_node_data);
  FREE(max_node_data);
  FREE(node_data);

  FREE(model_stats.node_comp_list);
  FREE(model_stats.node_active_list);

  FREE(min_cell_data);
  FREE(max_cell_data);
  FREE(cell_data);

  FREE(model_stats.cell_comp_list);
  FREE(model_stats.cell_active_list);

  FREE(min_model_data);
  FREE(max_model_data);
  FREE(model_data);

  FREE(model_stats.model_comp_list);
  FREE(model_stats.model_active_list);

  return 1;
}

Ctype* GetCells()
{
	return cells;
}

void SetCells(Ctype* p_cells)
{
	cells = p_cells;
}

Ntype* GetCellNlists()
{
	return cell_nlists;
}

void SetCellNlists(Ntype* p_cell_nlists)
{
	cell_nlists = p_cell_nlists;
}

Mtype* GetModelStats()
{
	return &model_stats;
}

void SetModelStats(Mtype p_model_stats)
{
	model_stats = p_model_stats;
}

float* GetXc()
{
	return xc;
}

void SetXc(float* t_xc)
{
	xc = t_xc;
}

float* GetYc()
{
	return yc;
}

void SetYc(float* t_yc)
{
	yc = t_yc;
}

float* GetZc()
{
	return zc;
}

void SetZc(float* t_zc)
{
	zc = t_zc;
}

float* GetNodeData()
{
	return node_data;
}

void SetNodeData(float* t_node_data)
{
	node_data = t_node_data;
}

float* GetCellData()
{
	return cell_data;
}

void SetCellData(float* t_cell_data)
{
	cell_data = t_cell_data;
}

float* GetModelData()
{
	return model_data;
}

void SetModelData(float* t_model_data)
{
	model_data = t_model_data;
}

float* GetMinNodeData()
{
	return min_node_data;
}

void SetMinNodeData(float* t_min_node_data)
{
	min_node_data = t_min_node_data;
}

float* GetMaxNodeData()
{
	return max_node_data;
}

void SetMaxNodeData(float* t_max_node_data)
{
	max_node_data = t_max_node_data;
}

float* GetMinCellData()
{
	return min_cell_data;
}

void SetMinCellData(float* t_min_cell_data)
{
	min_cell_data = t_min_cell_data;
}

float* GetMaxCellData()
{
	return max_cell_data;
}

void SetMaxCellData(float* t_max_cell_data)
{
	max_cell_data = t_max_cell_data;
}

float* GetMinModelData()
{
	return min_model_data;
}

void SetMinModelData(float* t_min_model_data)
{
	min_model_data = t_min_model_data;
}

float* GetMaxModelData()
{
	return max_model_data;
}

void SetMaxModelData(float* t_max_model_data)
{
	max_model_data = t_max_model_data;
}

/*  fin de la declaration des informations lues sur le fichier UCD   */

/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *             ****  readucd  :  read a ucd ascii file. ****                *
 *                                                                          *
 *--------------------------------------------------------------------------*/

static int   ascii = 1;

int readucd(FILE* file_num)
{
	int   ret = 0;

	/*  check to see if file is binary or ascii.  */

	if (check_file(file_num))
	{
	    ascii = 1;  /* this is an ASCII file */

	    if (!read_ascii (file_num, &cells, &cell_nlists, &model_stats,
			     &xc, &yc, &zc,
			     &node_data, &cell_data, &model_data,
			     &min_node_data, &max_node_data,
			     &min_cell_data, &max_cell_data,
			     &min_model_data, &max_model_data))
	    {
		fprintf(stderr,
		       " Error in read_ucd (ASCII format): pas de lecture %p\n",
		       file_num);
		ret = 1 ;
	    }

	}
	else
	{
	    int *bcell_nlists;
	    int cell;
	    int *pcell_nlists ;

	    ascii = 0;  /* this is not an ASCII file */

	    if (!read_binary (file_num, &cells, &bcell_nlists, &model_stats,
			      &xc, &yc, &zc,
			      &node_data, &cell_data, &model_data,
			      &min_node_data, &max_node_data,
			      &min_cell_data, &max_cell_data,
			      &min_model_data, &max_model_data))
	    {
		fprintf(stderr,
		      " Error in read_ucd (BINARY format): can't open file %p\n",
		      file_num);
		ret = 2 ;
	    }

	    cell_nlists  = MALLOC(Ntype,model_stats.num_cells);
	    pcell_nlists = bcell_nlists;

	    for ( cell=0; cell< model_stats.num_cells; cell++)
	    {
		int n = cells[cell].n;

		int node;

		assert( bcell_nlists != NULL );

		cell_nlists[cell].n = n;

		cell_nlists[cell].node_list =
		  MALLOC( int,num_node[cells[cell].cell_type] );
		for (node=0; node<n; node++ )
		{
		    cell_nlists[cell].node_list[node] = *(pcell_nlists++);
		}
		
	    }

	    free(bcell_nlists);
	}

	/* verification du maillage */
	if ( 0 )
	{
	    int cell;
	    int num_nodes = model_stats.num_nodes;
	    int num_cells = model_stats.num_cells;
	    int *n1, *n2, *n3, *n4;
	    int *n5, *n6, *n7, *n8;
	    float *volelt;

	    int *n[8];

	    fprintf(stderr," %p\n\t num_nodes = %d num_cells = %d\n",
		    file_num, num_nodes, num_cells);

	    n[0] = n1 = MALLOC( int, num_cells );
	    n[1] = n2 = MALLOC( int, num_cells );
	    n[2] = n3 = MALLOC( int, num_cells );
	    n[3] = n4 = MALLOC( int, num_cells );
	    n[4] = n5 = MALLOC( int, num_cells );
	    n[5] = n6 = MALLOC( int, num_cells );
	    n[6] = n7 = MALLOC( int, num_cells );
	    n[7] = n8 = MALLOC( int, num_cells );

	    /*fprintf(stderr," Connectivity table :\n"); */

	    for ( cell=0; cell< num_cells; cell++)
	    {
		int node;

		for (node=0; node<cells[cell].n; node++ )
		{
		/*    fprintf(stderr," %d",cell_nlists[cell].node_list[node]); */
		    n[node][cell] = cell_nlists[cell].node_list[node];
		}

		if ( node<cells[cell].n && node < 7 )
		{
		    int end = n[ cells[cell].n - 1 ][cell];

		    for (node=cells[cell].n; node < 8; node++ )
		    {
			n[node][cell] = end;
		    }

		}
		/*  fprintf(stderr,"\n");  */
	    }

	    volelt = MALLOC( float, num_cells );

	    l3volum(num_cells, num_nodes, xc, yc,zc,
		    n1, n2, n3, n4, n5, n6, n7, n8, volelt);
	    
	    for ( cell=0; cell< num_cells; cell++)
	    {

		/*
		   if ( (float) fabs( (double)volelt[cell] ) < FLT_EPSILON )
		   */
		if ( volelt[cell] < FLT_EPSILON )
		{
		    fprintf(stderr,"-WARNING- cell %d has an odd volume <%e>\n",
			    cell, volelt[cell]);
		    fprintf(stderr,
			    "\tnode_list[0:%d] = %d %d %d %d %d %d %d %d\n",
			    cells[cell].n,
			    n[0][cell], n[1][cell], n[2][cell], n[3][cell],
			    n[4][cell], n[5][cell], n[6][cell], n[7][cell]);
		}

	    }

	}

	return ret;
}


/*
 *--------------------------------------------------------------------------*
 *                                                                          *
 *             ****  printucd  :  print an ucd ascii file. ****             *
 *                                                                          *
 *--------------------------------------------------------------------------*/

int printucd(FILE* file_num,int choix_ascii)
{
	int   ret = 0;

	ascii = choix_ascii;

	/*  check to see if file is binary or ascii.  */

	if ( ascii )
	{
	    printf(" %p ASCII file\n",file_num);

	    write_ascii (file_num, cells, cell_nlists, model_stats, xc, yc, zc,
			 node_data, cell_data, model_data,
			 min_node_data, max_node_data,
			 min_cell_data, max_cell_data,
			 min_model_data, max_model_data);
	}
	else
	{
	    int num_nlist_nodes = model_stats.num_nlist_nodes;
	    int *bcell_nlists = MALLOC( int,num_nlist_nodes );
	    int *pcell_nlists = bcell_nlists;
	    int cell;

	    for ( cell=0; cell< model_stats.num_cells; cell++)
	    {
		int n = cells[cell].n;

		int node;

		for (node=0; node<n; node++ )
		{
		    *(pcell_nlists++) = cell_nlists[cell].node_list[node];
		}

	    }

	    printf(" %p BINARY file\n",file_num);

	    write_binary (file_num, cells, bcell_nlists, model_stats, xc, yc, zc,
			  node_data, cell_data, model_data,
			  min_node_data, max_node_data,
			  min_cell_data, max_cell_data,
			  min_model_data, max_model_data);

	    free(bcell_nlists);
	}

	return ret;
}

#ifdef TEST_UCD

#include "ucd.h"

int
main(int argc, char *argv[])
{
	int ret = 0;

	/* Programme de teste "bidon", on lit un fichier ucd ( xx.inp ) indique
	   comme deuxieme argument et on genere un fichier ucd (yy.inp ) indique
	   comme troisieme argument.

	   Compilation : cc -DTEST_UCD -g ucd.c -o ucd
	   Execution   : ucd avs.inp /tmp/avs.inp

	   Ou : cc = cc -ansiposix sur SGI 
	   cc = acc -Xc sur SUN
	   */

	if ( (argc > 3) | (argc < 2) )
	{
	    fprintf(stderr," Error : %s needs two files name argument\n",
		    argv[0]);
	    ret = 2;
	}
	else if (argc==2)
	{
	    ret = readucd( argv[1] );
	}
	else /* if (argc==3) */
	{
	    ret = readucd( argv[1] );
	    if ( ! ret ) ret = printucd( argv[2],!ascii);
	}

	return ret;
}
#endif /* fin TEST_UCD */
