/*******************/
/* L3Fonctions3D.c */
/*******************/

#include <stdio.h>
#include <math.h>

#include "L3Environnement.h"
#include "L3Fonctions3D.h"

#ifndef lint
    static const char sccs_id[] = "%W% %G% DMA/AIM/LM";
#endif

/*************************************************************************************************
 * Procedure utilisateur : L3Projection2D                                                        *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage   : numero de la sous-image concernee                                             *
 *   x3D,y3D,z3D : coordonnees du point de l'espace 3D a projeter                                *
 *   p_x2D,p_y2D : pointeurs sur les coordonnees du point dans le plan de projection             *
 *   p_position  : pointeur sur la variable devant indiquer si le point est visible ou invisible *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs :                                                                                     *
 *   espace 3D non defini                                                                        *
 *************************************************************************************************/


void
L3Projection2D(int sousimage, REEL x3D, REEL y3D, REEL z3D, REEL * p_x2D, REEL * p_y2D, int *p_position)
{
	double          x, y, z;
	double          xt, yt, zt;

	SETSOUSIMAGE(sousimage);

	if (VISU3D == VRAI) {

		/*-----------------------------------------------------------------*
		 * On se ramene a un repere centre sur l'oeil, dont l'axe vertical *
		 * passe par l'oeil et le point vise et oriente du point vise vers *
		 * l'oeil. Les coordonnees du point vise dans ce repere sont :     *
		 * (0.0, 0.0, -R) ou R est la distance oeil-point vise. Le plan de *
		 * projection est donc le plan d'equation z=-R dans ce repere.     *
		 *-----------------------------------------------------------------*/

		/*-------------*/
		/* Translation */
		/*-------------*/

		xt = x3D - XOEIL;
		yt = y3D - YOEIL;
		zt = z3D - ZOEIL;

		/*----------------------*/
		/* Changement de repere */
		/*----------------------*/

		x = COEFXX * xt + COEFXY * yt + COEFXZ * zt;
		y = COEFYX * xt + COEFYY * yt + COEFYZ * zt;
		z = COEFZX * xt + COEFZY * yt + COEFZZ * zt;

		if (z >= 0.0)
			*p_position = DERRIEREOEIL;
		else {
			*p_position = DEVANTOEIL;

			/*------------------------------------------------------------*
			 * Projection conique correspondant a une homothetie centree  *
			 * sur l'oeil et de rapport -R/z (positif dans ce repere pour *
			 * les points visibles, soit avec z<0.0).                     *
			 * Le facteur COEF3D permet de se ramener a un intervalle     *
			 * [-1.0,+1.0] pour les coordonnees projetees obtenues.       *
			 *------------------------------------------------------------*/

			*p_x2D = -COEF3D * x * R / z;
			*p_y2D = -COEF3D * y * R / z;
		}

	} else {
		/*--------------------------------*/
		/* Erreur : espace 3D non defini. */
		/*--------------------------------*/
		(void) fprintf(stderr, "L3Projection2D : espace 3D non defini.\n");
	}
}


/*************************************************************************************************
 * Procedure utilisateur : L3DefSpace3D                                                          *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage         : numero de la sous-image concernee                                       *
 *   xoeil,yoeil,zoeil : coordonnees de l'oeil dans l'espace 3D                                  *
 *   xvise,yvise,zvise : coordonnees du point vise dans l'espace 3D                              *
 *   angle             : angle de champ en radians                                               *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs :                                                                                     *
 *   espace 3D incorrect                                                                         *
 *************************************************************************************************/


void
L3DefSpace3D(int sousimage, REEL xoeil, REEL yoeil, REEL zoeil, REEL xvise, REEL yvise, REEL zvise, double angle)
{
	double          xvo, yvo, zvo;
	double          distxy;

	if (((xoeil != xvise) || (yoeil != yvise) || (zoeil != zvise))
	    && (angle > 0.0) && (angle < L3Pi)) {

		L3DefSpace(sousimage, -1.0, -1.0, 1.0, 1.0);

		/*-----------------------------------------------------------*/
		/* L3DefSpace selectionne la sous-image et met VISU3D a FAUX */
		/*-----------------------------------------------------------*/

		VISU3D = VRAI;

		XOEIL = xoeil;
		YOEIL = yoeil;
		ZOEIL = zoeil;

		xvo = xoeil - xvise;
		yvo = yoeil - yvise;
		zvo = zoeil - zvise;
		distxy = sqrt(xvo * xvo + yvo * yvo);

		/*---------------------------------------*/
		/* Calcul de la distance oeil-point vise */
		/*---------------------------------------*/

		R = sqrt(distxy * distxy + zvo * zvo);

		/*--------------------------------------------------*
		 * Calcul du coefficient permettant de se ramener a *
		 * un espace de coordonnees allant de -1.0 a +1.0.  *
		 *--------------------------------------------------*/

		COEF3D = 1.0 / tan(angle / 2.0) / R;

		/*------------------------------------------*
		 * Calcul des coefficients de la matrice de *
		 * changement de repere.                    *
		 *------------------------------------------*/

		if (distxy == 0.0) {
			/*--------------------*/
			/* Matrice identite : */
			/*--------------------*/
			COEFXX = 1.0;
			COEFXY = 0.0;
			COEFXZ = 0.0;
			COEFYX = 0.0;
			COEFYY = 1.0;
			COEFYZ = 0.0;
			COEFZX = 0.0;
			COEFZY = 0.0;
			COEFZZ = 1.0;
		} else {
			COEFXX = -yvo / distxy;
			COEFXY = xvo / distxy;
			COEFXZ = 0.0;

			COEFYX = -xvo * zvo / R / distxy;
			COEFYY = -yvo * zvo / R / distxy;
			COEFYZ = distxy / R;

			COEFZX = xvo / R;
			COEFZY = yvo / R;
			COEFZZ = zvo / R;
		}
	} else {
		/*------------------------------------------------*/
		/* Erreur : definition incorrecte de l'espace 3D. */
		/*------------------------------------------------*/
		(void) fprintf(stderr, "L3DefSpace3D : definition incorrecte de l'espace 3D.\n");
	}
}


/*************************************************************************************************
 * Procedure utilisateur : L3PointCourant3D                                                      *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage : numero de la sous-image concernee                                               *
 *   x,y,z     : coordonnees du nouveau point courant dans l'espace 3D                           *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs : aucune                                                                              *
 *************************************************************************************************/


void
L3PointCourant3D(int sousimage, REEL x, REEL y, REEL z)
{
	REEL            x2D, y2D;
	int             position;

	L3Projection2D(sousimage, x, y, z, &x2D, &y2D, &position);
	L3PointCourant(sousimage, x2D, y2D);
}


/*************************************************************************************************
 * Procedure utilisateur : L3LigneBrisee3D                                                       *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage : numero de la sous-image concernee                                               *
 *   x,y,z     : tableaux de coordonnees de la ligne brisee dans l'espace 3D                     *
 *   nlignes   : nombre de segments dans la ligne brisee                                         *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs :                                                                                     *
 *   nombre de segments negatif ou nul                                                           *
 *************************************************************************************************/


void
L3LigneBrisee3D(sousimage, x, y, z, nlignes)
	int             sousimage;
	REEL            x[], y[], z[];
	int             nlignes;

{
	REEL            x2D, y2D;
	int             position;
	int             i;

	if (nlignes > 0) {
		for (i = 0; i < nlignes; i++) {
			L3Projection2D(sousimage, x[i], y[i], z[i], &x2D, &y2D, &position);
			if (position != DERRIEREOEIL)
				L3LigneBrisee(sousimage, &x2D, &y2D, (int) 1);
			else
				L3PointCourant(sousimage, x2D, y2D);
		}
	} else {
		/*-------------------------------------------*/
		/* Erreur : nombre de lignes negatif ou nul. */
		/*-------------------------------------------*/
		(void) fprintf(stderr, "L3LigneBrisee3D : nombre de lignes negatif ou nul :%d\n", nlignes);
	}
}


/*************************************************************************************************
 * Procedure utilisateur : L3Segment3D                                                           *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage : numero de la sous-image concernee                                               *
 *   x1,y1,z1  : tableaux de coordonnees du debut des segments de l'espace 3D                    *
 *   x2,y2,z2  : tableaux de coordonnees de la fin des segments de l'espace 3D                   *
 *   nsegments : nombre de segments a tracer                                                     *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs :                                                                                     *
 *   nombre de segments negatif ou nul                                                           *
 *************************************************************************************************/


void
L3Segment3D(sousimage, x1, y1, z1, x2, y2, z2, nsegments)
	int             sousimage;
	REEL            x1[], y1[], z1[];
	REEL            x2[], y2[], z2[];
	int             nsegments;

{
	REEL            x2D1, y2D1;
	REEL            x2D2, y2D2;
	int             position1, position2;
	int             i;

	if (nsegments > 0) {
		for (i = 0; i < nsegments; i++) {
			L3Projection2D(sousimage, x1[i], y1[i], z1[i], &x2D1, &y2D1, &position1);
			L3Projection2D(sousimage, x2[i], y2[i], z2[i], &x2D2, &y2D2, &position2);
			if ((position1 != DERRIEREOEIL) && (position2 != DERRIEREOEIL))
				L3Segment(sousimage, &x2D1, &y2D1, &x2D2, &y2D2, (int) 1);
		}
	} else {
		/*---------------------------------------------*/
		/* Erreur : nombre de segments negatif ou nul. */
		/*---------------------------------------------*/
		(void) fprintf(stderr, "L3Segment3D : nombre de segments negatif ou nul :%d\n", nsegments);
	}
}


/*************************************************************************************************
 * Procedure utilisateur : L3Marqueur3D                                                          *
 *-----------------------------------------------------------------------------------------------*
 * Parametres :                                                                                  *
 *   sousimage    : numero de la sous-image concernee                                            *
 *   x,y,z        : tableaux de coordonnees des points de l'espace 3D a marquer                  *
 *   npoints      : nombre de points a marquer                                                   *
 *   typemarqueur : type du marqueur a tracer                                                    *
 *                                                                                               *
 * Valeur de retour : aucune                                                                     *
 *                                                                                               *
 * Description : voir document CEL.V DMA/SIE/IG NT 58.9 de juillet 1987 "Graf CIGALE"            *
 *                                                                                               *
 * Erreurs :                                                                                     *
 *   nombre de points negatif ou nul                                                             *
 *************************************************************************************************/


void
L3Marqueur3D(sousimage, x, y, z, npoints, typemarqueur)
	int             sousimage;
	REEL            x[], y[], z[];
	int             npoints;
	int             typemarqueur;

{
	REEL            x2D, y2D;
	int             position;
	int             i;

	if (npoints > 0) {
		for (i = 0; i < npoints; i++) {
			L3Projection2D(sousimage, x[i], y[i], z[i], &x2D, &y2D, &position);
			if (position != DERRIEREOEIL)
				L3Marqueur(sousimage, &x2D, &y2D, (int) 1, typemarqueur);
		}
	} else {
		/*-------------------------------------------*/
		/* Erreur : nombre de points negatif ou nul. */
		/*-------------------------------------------*/
		(void) fprintf(stderr, "L3Marqueur3D : nombre de points negatif ou nul :%d\n", npoints);
	}
}
